#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import datetime
import itertools
from typing import Dict, List
from unittest.mock import MagicMock, Mock

import pytest
from core_sampler.database import PGRowSet
from core_sampler.interfaces import Relationship, RowSet, RowWriter, Table
from core_sampler.row_writer import (
    PostgresCopyRowWriter,
    TopologicallySortingRowWriter,
    UniquifyingRowWriter,
)


@pytest.fixture
def project_rowset() -> RowSet:
    projects = MagicMock(Table)
    projects.name = "projects"

    row = dict(
        id=0,
        name="this",
        timestamp=datetime.datetime.fromtimestamp(1626473810.01041),
        existential_crisis=None,
    )

    return PGRowSet(projects, [row])


def test_copy_output(capsys, project_rowset: RowSet):
    """
    Are we formatting the output properly for PostgreSQL?
    """
    rw = PostgresCopyRowWriter()
    project_rowset.write_to(rw)
    lines = capsys.readouterr().out.strip().split("\n")
    assert len(lines) == 3
    assert lines[0] == "COPY projects (id, name, timestamp, existential_crisis) FROM stdin;"
    assert lines[1] == "0	this	2021-07-16T16:16:50.010410	\\N"
    assert lines[2] == "\\."


def test_uniquifier(capsys, project_rowset: RowSet):
    fake_writer = MagicMock(RowWriter)
    uniq = UniquifyingRowWriter(fake_writer)

    # write the rows twice
    project_rowset.write_to(uniq)
    project_rowset.write_to(uniq)

    # we should only get called once though
    assert fake_writer.write_rows.call_count == 1


def test_topological_sort(capsys):
    # The problem here can probably arise with just two tables, but I want to show it with three.
    # The basic idea is that if A -> B and B -> C and A -> C, then the topological sort is A, B, C.
    # But that order must be respected no matter what order we happen to find rows we want to output.
    # So to prove it works, we're going to generate a single row in three tables, A, B, and C with
    # the relationships as described above, and then try outputting in each permutation, and make
    # sure that we get them in ABC order regardless of what order permutation we choose.

    a = MagicMock(Table)
    a.name = "a"
    a.rowset = PGRowSet(a, [dict(name="a")])

    b = MagicMock(Table)
    b.name = "b"
    b.rowset = PGRowSet(b, [dict(name="b")])

    c = MagicMock(Table)
    c.name = "c"
    c.rowset = PGRowSet(c, [dict(name="c")])

    def make_relationship(source, dest):
        r = MagicMock(Relationship)
        r.source_table = source
        r.destination_table = dest
        return r

    a.relations = Mock("relationships")
    b.relations = Mock("relationships")
    c.relations = Mock("relationships")

    a.relations.return_value = [make_relationship(a, b), make_relationship(a, c)]
    b.relations.return_value = [make_relationship(b, c)]
    c.relations.return_value = []

    class TableNameCapturer(RowWriter):
        def __init__(self):
            self.tables = []

        def write_rows(self, table: Table, rows: List[Dict]):
            # we aren't worried about the rows for the purposes of the test
            self.tables.append(table.name)

    # for every permutation of these things, I want to see the same output order of a, b, c
    for t1, t2, t3 in itertools.permutations([a, b, c]):
        capturer = TableNameCapturer()
        writer = TopologicallySortingRowWriter(capturer)
        writer.write_rows(t1, t1.rowset)
        writer.write_rows(t2, t2.rowset)
        writer.write_rows(t3, t3.rowset)
        writer.close()
        assert capturer.tables == ["a", "b", "c"]
